import pathlib
import threading
import timeit

import hashToSchema
import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    INPUT_CHANNEL,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_STRING_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    MetricPrefix,
    PythonDevice,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES
from karabo.common.states import State

from . import scenes, shmem_utils, utils
from ._version import version as deviceVersion


PROCESSING_STATE_TIMEOUT = 10


@KARABO_CLASSINFO("BaseCorrection", deviceVersion)
class BaseCorrection(PythonDevice):
    _correction_flag_class = None  # subclass must set to some enum class
    _gpu_runner_class = None  # subclass must set this
    _gpu_runner_init_args = {}  # subclass can set this (TODO: remove, design better)
    _managed_keys = [
        "outputShmemBufferSize",
        "dataFormat.outputAxisOrder",
        "dataFormat.outputImageDtype",
        "preview.enable",
        "preview.pulse",
        "preview.trainIdModulo",
        "corrections.disableAll",
        "loadMostRecentConstants",
    ]  # subclass must extend this and put it in schema
    _schema_cache_fields = {
        "doAnything",
        "dataFormat.memoryCells",
        "constantParameters.memoryCells",
        "dataFormat.pixelsX",
        "dataFormat.pixelsY",
        "dataFormat.outputAxisOrder",
        "preview.enable",
        "preview.pulse",
        "preview.trainIdModulo",
        "processingStateTimeout",
        "state",
    }  # subclass should be aware of cache, but does not need to extend

    def _load_constant_to_gpu(constant_name, constant_data):
        raise NotImplementedError()

    @property
    def input_data_shape(self):
        raise NotImplementedError()

    @property
    def output_data_shape(self):
        axis_lengths = {
            "x": self._schema_cache["dataFormat.pixelsX"],
            "y": self._schema_cache["dataFormat.pixelsY"],
            "c": self._schema_cache["dataFormat.memoryCells"],
        }
        return tuple(
            axis_lengths[axis]
            for axis in self._schema_cache["dataFormat.outputAxisOrder"]
        )

    @staticmethod
    def expectedParameters(expected):
        (
            INPUT_CHANNEL(expected).key("dataInput").commit(),
            # note: output schema not set, will be updated to match data later
            OUTPUT_CHANNEL(expected).key("dataOutput").commit(),
            VECTOR_STRING_ELEMENT(expected)
            .key("fastSources")
            .displayedName("Fast data sources")
            .description(
                "Sources to get data from. Only incoming hashes from these sources "
                "will be processed."
            )
            .assignmentOptional()
            .defaultValue([])
            .commit(),
            STRING_ELEMENT(expected)
            .key("pulseFilter")
            .displayedName("[disabled] Pulse filter")
            .description(
                "Filter pulses: will be evaluated as array of indices to keep from "
                "data. Can be anything which can be turned into numpy uint16 array. "
                "Numpy is available as np. Take care not to include duplicates. If "
                "empty, will not filter at all."
            )
            .readOnly()
            .initialValue("")
            .commit(),
            UINT32_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .displayedName("Output buffer size limit (GB)")
            .description(
                "Corrected trains are written to shared memory locations. These are "
                "pre-allocated and re-used. This parameter determines how big (number "
                "of GB) the circular buffer will be."
            )
            .assignmentOptional()
            .defaultValue(10)
            .commit(),
            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview", "constants"])
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("dataFormat")
            .displayedName("Data format (in/out)")
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.inputImageDtype")
            .displayedName("Input image data dtype")
            .description("The (numpy) dtype to expect for incoming image data.")
            .options("uint16,float32")
            .assignmentOptional()
            .defaultValue("uint16")
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.outputImageDtype")
            .displayedName("Output image data dtype")
            .description(
                "The (numpy) dtype to use for outgoing image data. Input is "
                "cast to float32, corrections are applied, and only then will "
                "the result be cast back to outputImageDtype (all on GPU)."
            )
            .options("float16,float32,uint16")
            .assignmentOptional()
            .defaultValue("float32")
            .commit(),
            # important: shape of data as going into correction
            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .displayedName("Pixels x")
            .description("Number of pixels of image data along X axis")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .displayedName("Pixels y")
            .description("Number of pixels of image data along Y axis")
            .assignmentOptional()
            .defaultValue(128)
            .commit(),
            UINT32_ELEMENT(expected)
            .key("dataFormat.memoryCells")
            .displayedName("Memory cells")
            .description("Full number of memory cells in incoming data")
            .assignmentOptional()
            .noDefaultValue()
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .displayedName("Output axis order")
            .description(
                "Axes of main data output can be reordered after correction. Axis order "
                "is specified as string consisting of 'x', 'y', and 'c', with the "
                "latter indicating the memory cell axis. The default value of 'cxy' "
                "puts pixels along the fast axes."
            )
            .assignmentOptional()
            .defaultValue("cxy")
            .commit(),
            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.inputDataShape")
            .displayedName("Input data shape")
            .description(
                "Image data shape in incoming data (from reader / DAQ). This value is "
                "computed from pixelsX, pixelsY, and memoryCells - this field just "
                "shows you what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),
            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.outputDataShape")
            .displayedName("Output data shape")
            .description(
                "Image data shape for data output from this device. This value is "
                "computed from pixelsX, pixelsY, and the size of the pulse filter - "
                "this field just shows what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),
        )

        (
            SLOT_ELEMENT(expected)
            .key("loadMostRecentConstants")
            .displayedName("Load most recent constants")
            .commit()
        )

        preview_schema = Schema()
        (
            NODE_ELEMENT(expected).key("preview").displayedName("Preview").commit(),
            NODE_ELEMENT(preview_schema).key("data").commit(),
            NDARRAY_ELEMENT(preview_schema).key("data.adc").dtype("FLOAT").commit(),
            OUTPUT_CHANNEL(expected)
            .key("preview.outputRaw")
            .dataSchema(preview_schema)
            .commit(),
            OUTPUT_CHANNEL(expected)
            .key("preview.outputCorrected")
            .dataSchema(preview_schema)
            .commit(),
            BOOL_ELEMENT(expected)
            .key("preview.enable")
            .displayedName("Enable preview data generation")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            INT32_ELEMENT(expected)
            .key("preview.pulse")
            .displayedName("Pulse (or stat) for preview")
            .description(
                "If this value is ≥ 0, the corresponding index from data will be "
                "sliced for the preview. If this value is ≤ 0, preview will be one of "
                "the following stats:"
                "-1: max, "
                "-2: mean, "
                "-3: sum, "
                "-4: stdev. "
                "Max means selecting the pulse with the maximum integrated value. The "
                "others are computed across all filtered pulses in the train."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(expected)
            .key("preview.trainIdModulo")
            .displayedName("Train modulo for throttling")
            .description(
                "Preview will only be generated for trains whose ID modulo this "
                "number is zero. Higher values means fewer preview updates. Should be "
                "adjusted based on input rate. Keep in mind that the GUI has limited "
                "refresh rate anyway and that network is precious."
            )
            .assignmentOptional()
            .defaultValue(6)
            .reconfigurable()
            .commit(),
        )

        # just measurements and counters to display
        (
            UINT64_ELEMENT(expected)
            .key("trainId")
            .displayedName("Train ID")
            .description("ID of latest train processed by this device.")
            .readOnly()
            .initialValue(0)
            .commit(),
            NODE_ELEMENT(expected)
            .key("performance")
            .displayedName("Performance measures")
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("performance.processingDuration")
            .displayedName("Processing time")
            .description(
                "Exponential moving average over time spent processing individual "
                "trains Time includes generating preview and sending data."
            )
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(100)
            .info("Processing not fast enough for full speed")
            .needsAcknowledging(False)
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("performance.rate")
            .displayedName("Rate")
            .description(
                "Actual rate with which this device gets, processes, and sends trains. "
                "This is a simple moving average."
            )
            .unit(Unit.HERTZ)
            .readOnly()
            .initialValue(0)
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("corrections")
            .displayedName("Correction steps")
            .commit(),
            BOOL_ELEMENT(expected)
            .key("corrections.disableAll")
            .displayedName("Disable corrections for dataOutput")
            .description(
                "Toggle for disabling all corrections for dataOutput at once. This "
                "overrides the individual flags under corrections.enabled (does not "
                "affect corrected preview). When corrections are disabled, the device "
                "still reshapes data to output shape, applies the pulse filter, and "
                "casts to output dtype. Disabling all corrections can be useful if "
                "constants are missing / bad, or if user software does not want "
                "corrected data."
            )
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
        )

    def __init__(self, config):
        self._schema_cache = {
            k: config.get(k) for k in self._schema_cache_fields if config.has(k)
        }
        super().__init__(config)

        if not sorted(config.get("dataFormat.outputAxisOrder")) == ["c", "x", "y"]:
            # TODO: figure out how to get this information to operator
            self.log_status_error("Invalid output axis order string")
            return

        self.KARABO_ON_DATA("dataInput", self.process_input)
        self.KARABO_ON_EOS("dataInput", self.handle_eos)

        self.sources = set(config.get("fastSources"))

        self.input_data_dtype = np.dtype(config.get("dataFormat.inputImageDtype"))
        self.output_data_dtype = np.dtype(config.get("dataFormat.outputImageDtype"))
        self.gpu_runner = None  # must call _update_shapes() in subclass init

        self.calcat_friend = self._calcat_friend_class(
            self, pathlib.Path.cwd() / "calibration-client-secrets.json"
        )

        self._correction_flag_enabled = self._correction_flag_class.NONE
        self._correction_flag_preview = self._correction_flag_class.NONE

        self._shmem_buffer = None
        self._has_set_output_schema = False
        self._has_updated_shapes = False
        self._processing_time_ema = utils.ExponentialMovingAverage(alpha=0.3)
        self._rate_tracker = utils.WindowRateTracker()

        self._buffered_status_update = Hash(
            "trainId",
            0,
            "performance.rate",
            0,
            "performance.processingDuration",
            0,
        )
        self._last_processing_started = 0  # not input handler should put timestamp
        self._rate_update_timer = utils.RepeatingTimer(
            interval=0.5,
            callback=self._update_rate_and_state,
        )
        self._buffer_lock = threading.Lock()
        self.KARABO_SLOT(self.loadMostRecentConstants)
        self.KARABO_SLOT(self.requestScene)

        # TODO: the CalCatFriend could add these for us
        # note: overly complicated slot function creation necessary for closure to work
        def make_wrapper_capturing_constant(constant):
            def aux():
                self.calcat_friend.get_specific_constant_version_and_call_me_back(
                    constant, self._load_constant_to_gpu
                )

            return aux

        for constant in self._constant_enum_class:
            slot_name = f"foundConstants.{constant.name}.overrideConstantVersion"
            meth_name = slot_name.replace(".", "_")
            self.KARABO_SLOT(
                make_wrapper_capturing_constant(constant),
                slotName=meth_name,
            )

    def preReconfigure(self, config):
        self._prereconfigure_update_hash = config

    def postReconfigure(self):
        if not hasattr(self, "_prereconfigure_update_hash"):
            self.log_status_warn("postReconfigure without knowing update hash")
            return

        update = self._prereconfigure_update_hash
        for path in update.getPaths():
            if path in self._schema_cache_fields:
                self._schema_cache[path] = update.get(path)

        # TODO: pulse filter (after reimplementing)
        if any(
            update.has(shape_param)
            for shape_param in (
                "dataFormat.pixelsX",
                "dataFormat.pixelsY",
                "dataFormat.memoryCells",
                "constantParameters.memoryCells",
            )
        ):
            self._update_shapes()
        # TODO: only call this if they are changed (is cheap, though)
        self._update_correction_flags()

    def loadMostRecentConstants(self):
        self.flush_constants()
        self.calcat_friend.flush_constants()
        for constant in self._constant_enum_class:
            self.calcat_friend.get_constant_version_and_call_me_back(
                constant, self._load_constant_to_gpu
            )

    def flush_constants(self):
        """Override from CalibrationReceiverBaseDevice to also flush GPU buffers"""
        for correction_step, _ in self._correction_field_names:
            self.set(f"corrections.{correction_step}.available", False)
        self.gpu_runner.flush_buffers()
        self._update_correction_flags()

    def log_status_info(self, msg):
        self.log.INFO(msg)
        self.set("status", msg)

    def log_status_warn(self, msg):
        self.log.WARN(msg)
        self.set("status", msg)

    def log_status_error(self, msg):
        self.set("status", msg)
        self.log.ERROR(msg)
        self.updateState(State.ERROR)

    def set(self, *args):
        """Wrapper around PythonDevice.set to enable caching "hot" schema elements"""
        if len(args) == 2:
            key, value = args
            if key in self._schema_cache_fields:
                self._schema_cache[key] = value
        super().set(*args)

    def requestScene(self, params):
        payload = Hash()
        scene_name = params.get("name", default="")
        payload["name"] = scene_name
        payload["success"] = True
        if scene_name == "overview":
            payload["data"] = scenes.correction_device_overview_scene(
                device_id=self.getInstanceId()
            )
        elif scene_name == "constants":
            payload["data"] = scenes.correction_device_constants_scene(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
                param_prefix="constantParameters",
                status_prefix="foundConstants",
            )
        else:
            payload["success"] = False
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def _write_output(self, data, old_metadata):
        metadata = ChannelMetaData(
            old_metadata.get("source"),
            Timestamp.fromHashAttributes(old_metadata.getAttributes("timestamp")),
        )

        if not self._has_set_output_schema:
            self.updateState(State.CHANGING)
            self._update_output_schema(data)
            self.updateState(State.PROCESSING)

        channel = self.signalSlotable.getOutputChannel("dataOutput")
        channel.write(data, metadata, False)
        channel.update()

    def _write_combiner_preview(self, data_raw, data_corrected, train_id, source):
        # TODO: take into account updated pulse table after pulse filter
        # TODO: send as ImageData (requires updated assembler)
        # TODO: allow sending *all* frames for commissioning (request: Jola)
        preview_hash = Hash()
        preview_hash.set("image.trainId", train_id)
        preview_hash.set("image.pulseId", self._schema_cache["preview.pulse"])

        # note: have to construct because setting .tid after init is broken
        timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        metadata = ChannelMetaData(source, timestamp)
        for channel_name, data in (
            ("preview.outputRaw", data_raw),
            ("preview.outputCorrected", data_corrected),
        ):
            preview_hash.set("data.adc", data[..., np.newaxis])
            channel = self.signalSlotable.getOutputChannel(channel_name)
            channel.write(preview_hash, metadata, False)
            channel.update()

    def _update_correction_flags(self):
        """Based on constants loaded and settings, update bit mask flags for kernel"""
        available = self._correction_flag_class.NONE
        enabled = self._correction_flag_class.NONE
        preview = self._correction_flag_class.NONE
        for field_name, flag in self._correction_field_names:
            if self.get(f"corrections.{field_name}.available"):
                available |= flag
            if self.get(f"corrections.{field_name}.enable"):
                enabled |= flag
            if self.get(f"corrections.{field_name}.preview"):
                preview |= flag
        enabled &= available
        preview &= available
        if self.get("corrections.disableAll"):
            enabled &= self._correction_flag_class.NONE
        self._correction_flag_enabled = enabled
        self._correction_flag_preview = preview
        self.log.DEBUG(f"Corrections for dataOutput: {str(enabled)}")
        self.log.DEBUG(f"Corrections for preview: {str(preview)}")

    def _update_output_schema(self, data):
        """Updates the schema of dataOutput based on data we want to send

        This should only be called once: when handling output for the first
        time, we update the schema to match the modified data we'd send.
        """

        # TODO: remove when switching to specifying output schema up-front
        self.log.INFO("Updating output schema")
        my_schema_update = Schema()
        data_schema = hashToSchema.HashToSchema(data).schema
        (
            OUTPUT_CHANNEL(my_schema_update)
            .key("dataOutput")
            .dataSchema(data_schema)
            .commit()
        )
        self.appendSchema(my_schema_update)
        self._has_set_output_schema = True

    def _update_shapes(self):
        """(Re)initialize buffers according to expected data shapes"""
        self.log.INFO("Updating shapes")
        # reflect the axis reordering in the expected output shape
        self.set("dataFormat.inputDataShape", list(self.input_data_shape))
        self.set("dataFormat.outputDataShape", list(self.output_data_shape))

        if self._shmem_buffer is None:
            shmem_buffer_name = self.getInstanceId() + ":dataOutput"
            memory_budget = self.get("outputShmemBufferSize") * 2 ** 30
            self.log.INFO(f"Opening new shmem buffer: {shmem_buffer_name}")
            self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
                memory_budget,
                self.output_data_shape,
                self.output_data_dtype,
                shmem_buffer_name,
            )
        else:
            self._shmem_buffer.change_shape(self.output_data_shape)

        self.gpu_runner = self._gpu_runner_class(
            self.get("dataFormat.pixelsX"),
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.memoryCells"),
            int(self.get("constantParameters.memoryCells")),
            input_data_dtype=self.input_data_dtype,
            output_data_dtype=self.output_data_dtype,
            **self._gpu_runner_init_args,
        )

        # TODO: lock to avoid dict changing size underneath us
        for (
            constant,
            data,
        ) in self.calcat_friend.cached_constants.items():
            self.log_status_info(f"Reload constant {constant}")
            self._load_constant_to_gpu(constant, data)

        self._has_updated_shapes = True

    def _update_rate_and_state(self):
        self._buffered_status_update.set("performance.rate", self._rate_tracker.get())
        self._buffered_status_update.set(
            "performance.processingDuration", self._processing_time_ema.get() * 1000
        )
        # trainId should be set on _buffered_status_update in input handler
        self.set(self._buffered_status_update)

        if (
            timeit.default_timer() - self._last_processing_started
            > PROCESSING_STATE_TIMEOUT
        ):
            if self.get("state") is State.PROCESSING:
                self.updateState(State.ON)

    def handle_eos(self, channel):
        self._has_set_output_schema = False
        self.updateState(State.ON)
        self.signalEndOfStream("dataOutput")


def add_correction_step_schema(schema, managed_keys, field_flag_mapping):
    for field_name, _ in field_flag_mapping:
        node_name = f"corrections.{field_name}"
        (
            NODE_ELEMENT(schema).key(node_name).commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.available")
            .displayedName("Available")
            .description(
                "This boolean indicates whether the necessary constants have been "
                "loaded for this correction step to be applied. Enabling the "
                "correction will have no effect unless this is True."
            )
            .readOnly()
            .initialValue(False)
            .commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.enable")
            .displayedName("Enable")
            .description(
                "Controls whether to apply this correction step for main data "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.preview")
            .displayedName("Preview")
            .description(
                "Whether to apply this correction step for corrected preview "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )
        managed_keys.append(f"{node_name}.enable")
        managed_keys.append(f"{node_name}.preview")
